/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemLostException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.probes.AbsoluteResponseTimes;
import org.ow2.petals.probes.api.probes.DurationProbe;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.api.probes.StartDateItem;

/**
 * <p>
 * An abstract duration or response time probe to probe periodically durations or response times and storing:
 * </p>
 * <ul>
 * <li>as absolute values:
 * <ul>
 * <li>the max response time,</li>
 * <li>the avg response time,</li>
 * <li>the min response time,</li>
 * </ul>
 * </li>
 * <li>as relatives values:
 * <ul>
 * <li>the max response time,</li>
 * <li>the avg response time,</li>
 * <li>the min response time,</li>
 * <li>the 10-percentile response time: 10% of the response times are lesser than this value,</li>
 * <li>the 50-percentile response time: 50% of the response times are lesser than this value,</li>
 * <li>the 90-percentile response time: 90% of the response times are upper than this value.</li>
 * </ul>
 * </li>
 * </ul>
 * <p>
 * When the end of a sample occurs, relative response time values are computed with the response times collected during
 * the sample, and the absolute response time values are updated.
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u>The relative response time values can be acceded only if the probe is started. The absolute
 * response time values can be acceded even if the probe is stopped. The collection of response times starts when the
 * probe starts. The absolute and relative response time values are reseted when the probe starts.
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public abstract class AbstractDurationProbeImpl extends AbstractProbe implements DurationProbe {

    /**
     * The response time absolute values by key
     * 
     * No need for it to be concurrent because we always access it using {@link #absoluteResponseTimeValuesLock}!
     */
    private final Map<ExecutionStatus, AbsoluteResponseTimes> absoluteResponseTimeValues = new HashMap<ExecutionStatus, AbsoluteResponseTimes>(
            ExecutionStatus.values().length);

    private final ReadWriteLock absoluteResponseTimeValuesLock = new ReentrantReadWriteLock();

    /**
     * Timer used as sampler
     */
    private final Timer timer;

    /**
     * The periodic task switch collection samples. Recreated at each start/stop cycle to avoid PETALSCDK-116.
     */
    private TimerTask timerTask = null;

    /**
     * Duration of a sample in milliseconds.
     */
    private final long samplePeriod;

    /**
     * The previous sample used to access relative response time values during
     * the response time collection by <code>currentSample</code>
     */
    private ResponseTimesSample previousSample;

    /**
     * The sample collecting response times
     */
    private ResponseTimesSample currentSample;

    /**
     * Lock object to synchronize the samples switch.
     */
    private final ReadWriteLock sampleSwitchLock = new ReentrantReadWriteLock();

    /**
     * Creates a new keyed response time probe.
     * 
     * @param timer
     *            The timer used as sampler. <b>The caller is responsible to
     *            cancel the timer to free resources attached</b>.
     * @param samplePeriod
     *            The duration of a sample, in milliseconds.
     * @throws IllegalArgumentException
     *             <code>timer</code> is <code>null</code>, or
     *             <code>samplePeriod</code> has an invalid value.
     */
    public AbstractDurationProbeImpl(final Timer timer, final long samplePeriod)
            throws IllegalArgumentException {

        if (timer == null) {
            throw new IllegalArgumentException("The timer used as sampler is NULL !!");
        }

        if (samplePeriod <= 0) {
            throw new IllegalArgumentException("The sample period must strictly upper than 0");
        }

        // The previous and current sample are initialized on startup. see
        // 'doStart'

        this.timer = timer;
        this.samplePeriod = samplePeriod;
    }

    /**
     * Switches current and previous samples, updating absolute response time
     * values according to the current relative response time values.
     */
    private void switchesSamples() {

        // All following operation must be ATOMIC (no other operation on
        // response time must be done during the switch)
        this.absoluteResponseTimeValuesLock.writeLock().lock();
        try {
            this.sampleSwitchLock.writeLock().lock();
            try {
                this.currentSample.stopCollecting();
                this.previousSample = this.currentSample;
                this.currentSample = new ResponseTimesSample(this.previousSample);
            } finally {
                this.sampleSwitchLock.writeLock().unlock();
            }

            // From here, the current sample can be updated, data of the
            // previous sample can be returned, but ABSOLUTE DATA can
            // not be returned because they are not updated with
            // previous sample.
            final Iterator<Entry<ExecutionStatus, ResponseTimeRelativeValues>> itResponseTimeValues = this.previousSample
                    .getResponseTimeValues().entrySet().iterator();
            while (itResponseTimeValues.hasNext()) {
                final Entry<ExecutionStatus, ResponseTimeRelativeValues> responseTimeValuesEntry = itResponseTimeValues
                        .next();
                final ExecutionStatus key = responseTimeValuesEntry.getKey();
                final AbsoluteResponseTimes absoluteResponseTimeValues = this.absoluteResponseTimeValues
                        .get(key);
                if (absoluteResponseTimeValues != null && key != ExecutionStatus.PENDING) {
                    // Absolute values have no sens for pending requests
                    absoluteResponseTimeValues.update(responseTimeValuesEntry.getValue());
                } else {
                    this.absoluteResponseTimeValues.put(key, new AbsoluteResponseTimes(
                            responseTimeValuesEntry.getValue()));
                }
            }
        } finally {
            this.absoluteResponseTimeValuesLock.writeLock().unlock();
        }
    }

    /**
     * A new execution starts, invoke this method to pick the current time. When
     * the execution will end, use
     * {@link #endsExecution(StartDateItem, ExecutionStatus)} to close the
     * execution and compute the response time.
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     * @throws ResponseTimeCollectionStoppedException
     *             A sample switch is in progress, no response time collection
     *             is available. This exception should not occur.
     * @throws StartDateItemLostException
     *             Unable to store the start time of the new execution. This
     *             exception should not occur.
     * @return A {@link StartDateItem} created for this new execution.
     */
    public StartDateItem newExecution() throws ProbeNotStartedException,
            ResponseTimeCollectionStoppedException, StartDateItemLostException {
        final long startTime = this.getTimeStamp();
        if (this.isStarted) {
            this.sampleSwitchLock.readLock().lock();
            try {
                return this.currentSample.addStartDate(startTime);
            } finally {
                this.sampleSwitchLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Ends an execution and compute the response time.
     * 
     * @see #newExecution()
     * @param startDateItem
     *            The {@link StartDateItem} created by {@link #newExecution()}
     *            for the current execution.
     * @param executionStatus
     *            The execution status of this execution
     * @throws ProbeNotStartedException
     *             The probe is not started.
     * @throws ResponseTimeCollectionStoppedException
     * @throws StartDateItemUnknownException
     *             <code>startDateItem</code> not found in the list of start
     *             dates. This exception should not occur.
     */
    public void endsExecution(final StartDateItem startDateItem,
            final ExecutionStatus executionStatus) throws ProbeNotStartedException,
            ResponseTimeCollectionStoppedException, StartDateItemUnknownException {
        final long endTime = this.getTimeStamp();
        if (this.isStarted) {
            this.sampleSwitchLock.readLock().lock();
            try {
                this.currentSample.endStartDate(startDateItem, endTime, executionStatus);
            } finally {
                this.sampleSwitchLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Get the current timestamp. Each inherited class can have each own precision: milliseconds, nanoseconds, ...
     * 
     * @return The current timestamp.
     */
    protected abstract long getTimeStamp();

    /**
     * @return The absolute response time values.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public Map<ExecutionStatus, AbsoluteResponseTimes> getAbsoluteResponseTimeValues()
            throws ProbeNotInitializedException {

        if (this.isInitialized) {
            this.absoluteResponseTimeValuesLock.readLock().lock();
            try {
                // return a copy, if not it could be read or modified when we don't want to!
                return new HashMap<>(this.absoluteResponseTimeValues);
            } finally {
                this.absoluteResponseTimeValuesLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    /**
     * @return The absolute response time values converted into simple java
     *         types. Each key of the map entries is the execution status. Each
     *         value of the map entries is an array of long values such as:
     *         <ol>
     *         <li>the first long value is the max response time,</li>
     *         <li>the second long value is the average of response times,</li>
     *         <li>the third long value is the min response time.</li>
     *         </ol>
     * @throws ProbeNotInitializedException
     *             The probe is not initialized or shutdown.
     */
    public Map<String, Long[]> getConvertedAbsoluteResponseTimeValues()
            throws ProbeNotInitializedException {

        if (this.isInitialized) {
            this.absoluteResponseTimeValuesLock.readLock().lock();
            try {
                final Map<String, Long[]> results = new HashMap<>(absoluteResponseTimeValues.size());
                for (final Map.Entry<ExecutionStatus, AbsoluteResponseTimes> entry : absoluteResponseTimeValues
                        .entrySet()) {
                    final AbsoluteResponseTimes absoluteResponseTimesValues = entry.getValue();
                    results.put(entry.getKey().name(),
                            new Long[] { absoluteResponseTimesValues.getMaxResponseTime(),
                                    absoluteResponseTimesValues.getAvgResponseTime(),
                                    absoluteResponseTimesValues.getMinResponseTime() });
                }
                return results;
            } finally {
                this.absoluteResponseTimeValuesLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    /**
     * @return The relative response time values.
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public Map<ExecutionStatus, ResponseTimeRelativeValues> getRelativeResponseTimeValues()
            throws ProbeNotStartedException {

        if (this.isStarted) {
            this.sampleSwitchLock.readLock().lock();
            try {
                return this.previousSample.getResponseTimeValues();
            } finally {
                this.sampleSwitchLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * @return The relative response time values converted into simple java
     *         types. Each key of the map entries is the execution status. Each
     *         value of the map entries is an array of long values such as:
     *         <ol>
     *         <li>the first long value is the max response time,</li>
     *         <li>the second long value is the average of response times,</li>
     *         <li>the third long value is the min response time,</li>
     *         <li>the fourth long value is the 10-percentile response time,</li>
     *         <li>the fourth long value is the 50-percentile response time,</li>
     *         <li>the fourth long value is the 90-percentile response time.</li>
     *         </ol>
     * @throws ProbeNotStartedException
     *             The probe is not started or stopped.
     */
    public Map<String, Long[]> getConvertedRelativeResponseTimeValues()
            throws ProbeNotStartedException {

        if (this.isStarted) {
            this.sampleSwitchLock.readLock().lock();
            try {
                final Map<ExecutionStatus, ResponseTimeRelativeValues> previousTimes = this.previousSample
                        .getResponseTimeValues();
                final Map<String, Long[]> results = new HashMap<>(previousTimes.size());

                for (final Map.Entry<ExecutionStatus, ResponseTimeRelativeValues> entry : previousTimes.entrySet()) {
                    final ResponseTimeRelativeValues relativeResponseTimesValues = entry.getValue();
                    results.put(entry.getKey().name(),
                            new Long[] { relativeResponseTimesValues.getMax(),
                                    relativeResponseTimesValues.getAvg(),
                                    relativeResponseTimesValues.getMin(),
                                    relativeResponseTimesValues.getPercent10(),
                                    relativeResponseTimesValues.getPercent50(),
                                    relativeResponseTimesValues.getPercent90() });
                }
                return results;
            } finally {
                this.sampleSwitchLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    @Override
    public final void doInit() throws ProbeInitializationException {
        // NOP
    }

    @Override
    public final void doStart() throws ProbeStartupException {
        // Reset the previous and current samples
        this.previousSample = new ResponseTimesSample();
        this.previousSample.stopCollecting();
        this.currentSample = new ResponseTimesSample();

        // Start the sampler
        this.timerTask = new TimerTask() {
            @Override
            public void run() {
                AbstractDurationProbeImpl.this.switchesSamples();
            }
        };
        this.timer.schedule(this.timerTask, this.samplePeriod, this.samplePeriod);

        // Reset the absolute response time values
        this.absoluteResponseTimeValues.clear();
    }

    @Override
    public final void doStop() throws ProbeStopException {
        // Stop the sampler
        if (this.timerTask != null) {
            this.timerTask.cancel();
        }
        
        // merge relative response time values with absolute response time values
        this.switchesSamples();
    }

    @Override
    public final void doShutdown() throws ProbeShutdownException {
        // Free resources
        this.absoluteResponseTimeValues.clear();
    }
}
