/**
 * Copyright (c) 2014-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.ow2.petals.probes.api.exceptions.ProbeInitializationException;
import org.ow2.petals.probes.api.exceptions.ProbeNotInitializedException;
import org.ow2.petals.probes.api.exceptions.ProbeNotStartedException;
import org.ow2.petals.probes.api.exceptions.ProbeShutdownException;
import org.ow2.petals.probes.api.exceptions.ProbeStartupException;
import org.ow2.petals.probes.api.exceptions.ProbeStopException;
import org.ow2.petals.probes.api.probes.CounterGaugeProbe;

/**
 * <p>
 * A simple counter probe to count the number of something and storing the max reached value (that can be greater than
 * the max of instant values) since its last start.
 * </p>
 * <p>
 * <u><b>Lifecycle</b>:</u> The counter gauge value can be incremented or decremented only if the probe is started. The
 * instant value can be acceded only if the probe is started. The max and min values can be acceded even if the probe is
 * initialized or stopped. The max value is reseted when the probe starts.
 * </p>
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class CounterGaugeProbeImpl extends AbstractProbe implements CounterGaugeProbe {

    /**
     * The initial value of the counter
     */
    private final long initialValue;

    /**
     * Instant value of the counter
     */
    private long instantValue;

    /**
     * Max value of the counter
     */
    private long maxValue;

    /**
     * Min value of the counter
     */
    private long minValue;

    private final ReadWriteLock valuesLock = new ReentrantReadWriteLock();

    /**
     * Create a {@link CounterGaugeProbeImpl} initialized with the initial value '0'.
     */
    public CounterGaugeProbeImpl() {
        this(0);
    }

    /**
     * Create a {@link CounterGaugeProbeImpl} initialized with the given initial value.
     * 
     * @param initialValue
     *            The initial value of the counter gauge.
     */
    public CounterGaugeProbeImpl(final long initialValue) {
        this.initialValue = initialValue;
    }

    /**
     * Increments the value of the counter.
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    @Override
    public void inc() throws ProbeNotStartedException {
        if (this.isStarted) {
            valuesLock.writeLock().lock();
            try {
                // Increment the counter and check max value. The min value is not impacted by increment.
                if (++this.instantValue > this.maxValue) {
                    this.maxValue = this.instantValue;
                }
            } finally {
                valuesLock.writeLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Decrements the value of the counter.
     * 
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    @Override
    public void dec() throws ProbeNotStartedException {
        if (this.isStarted) {
            valuesLock.writeLock().lock();
            try {
                // Decrement the counter and check min value.The max value is not impacted by decrement.
                if (--this.instantValue < this.minValue) {
                    this.minValue = this.instantValue;
                }
            } finally {
                valuesLock.writeLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Accessor to the current counter value.
     * 
     * @return The current counter value.
     * @throws ProbeNotStartedException
     *             The probe is not started.
     */
    @Override
    public long getCurrentValue() throws ProbeNotStartedException {
        if (this.isStarted) {
            valuesLock.readLock().lock();
            try {
                return this.instantValue;
            } finally {
                valuesLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotStartedException();
        }
    }

    /**
     * Accessor to the max counter value.
     * 
     * @return The max counter value since the last probe start.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized.
     */
    @Override
    public long getMaxValue() throws ProbeNotInitializedException {
        if (this.isInitialized) {
            valuesLock.readLock().lock();
            try {
                return this.maxValue;
            }
 finally {
                valuesLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    /**
     * Accessor to the min counter value.
     * 
     * @return The min counter value since the last probe start.
     * @throws ProbeNotInitializedException
     *             The probe is not initialized.
     */
    @Override
    public long getMinValue() throws ProbeNotInitializedException {
        if (this.isInitialized) {
            valuesLock.readLock().lock();
            try {
                return this.minValue;
            } finally {
                valuesLock.readLock().unlock();
            }
        } else {
            throw new ProbeNotInitializedException();
        }
    }

    @Override
    public final void doInit() throws ProbeInitializationException {
        this.maxValue = initialValue;
        this.minValue = initialValue;
    }

    @Override
    public final void doStart() throws ProbeStartupException {
        // Reset the counter value
        this.instantValue = initialValue;
        this.maxValue = initialValue;
        this.minValue = initialValue;
    }

    @Override
    public final void doStop() throws ProbeStopException {
        // NOP
    }

    @Override
    public final void doShutdown() throws ProbeShutdownException {
        // NOP
    }
}
